! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: rotdet2sky_mod
!
! Authors:  Brendan Perry - NASA/GSFC (brendan.h.perry@nasa.gov)
!           comprehensive rewrite, modularization, for SAS-21
!           Steven L. Snowden - NASA/GSFC (original algorithm)
!           Kip D. Kuntz - NASA/JHU
!
! Description: (former) Extended Source Analysis Software
!
! Comments:
!
! $Id: rotdet2sky_mod.f90, v 1.3 2009/02/26 16:39:45 aibarra Exp $
!-
!
!  WAS: ESAS task rot_det_sky 
!  Rotates images in detector coordinates into sky coordinates provided
!  by a template image.  This was used in the XMM-ESAS package to rotate
!  the model particle background images, which are created by the programs
!  mosback/pnback/proton in detector coordinates, into sky coordinates. This
!  new version is modular (not only for ESAS uses). Support was added for
!  swcx images. 
!
!  INPUT:
!    Input template image (typically 900,900 Real32 or Int32), e.g. mos1S001-fovimt.fits
!    Input ESAS-produced image (780,780, usually Int32), e.g. mos1S001-bkgimdet-350-1100.fits
!    Input mask image (optional) (typically 900,900, Int32), e.g. mos1S001-maskimdet.fits
!    Output rotated image (900,900 Real32) (user selected name, or derived)
!    DETX central coord (from esky2det or user-input)
!    DETY central coord (from esky2det or user-input)
!    X reference pixel (template CRPIX1 or user-input)
!    Y reference pixel (template CRPIX1 or user-input)
!
! NEW example: rotdet2sky template=mos1S001-fovimt.fits \
!              inimage=mos1S001-improtdet-350-1100.fits \
!              outbkgimage=mos1S001-improtsky-350-1100.fits \
!              withmask=F withskyxy=T detx=1235.4 dety=-1629.0  \
!              withskyxy=F
!
!    OLD example: rot-back 1 1S001 350 1250 -1406.1 -1368.6 \
!                 450.9 450.9 mask.fits
!
!    Template image file name (hardwired to 900x900 pixels)
!    Detector image file name (hardwired to 780x780 pixels)
!
!  OUTPUT:
!    Tranformed image image (same size as the template)

module rotdet2sky_mod

  use dal
  use param 
  use errorHandling
  use sasdal
  use gofutils_mod  ! from goflib, formerly in marq_mod
  use gofmatrix_mod ! from goflib, formerly in marq_mod
  use gofio_mod     ! from goflib, formerly in subsn (SS)
  
  implicit none

  private

  character(len=*), parameter :: &
       tempimpar   = "intemplate", &  ! template image file
       inimpar     = "inimage", &     ! input image file to be rotated
       outimpar    = "outimage", &    ! output rotated image file 
       withmaskpar = "withmask", &    ! provide a mask file?
       inmaskpar   = "inmaskimage", & ! input optional mask image
       withdetpar  = "withdetxy", &   ! input detx/y or use crval1/2?
       detxcenpar  = "detx", &        ! input detx (from esky2det)
       detycenpar  = "dety", &        ! input dety (from esky2det)
       withskypar  = "withskyxy", &   ! input skyx/y or use crpix1/2?
       skyxpar     = "skyx", &        ! user input skyx/y if user
       skyypar     = "skyy"           ! chooses withskyxy=T (default)
  
  public rotdet2sky

contains

  !---------------------------------------------------------------------
  subroutine rotdet2sky
    !---------------------------------------------------------------------
    
    type(DataSetT) :: inTEMPIMG_S, inIMAGE_S, inMASKIMG_S, outIMAGE_S
    type(ArrayT)   :: inTEMPIMG_A, inIMAGE_A, inMASKIMG_A, outIMAGE_A
    type(TableT)   :: in_T
    !
    character(len=4) :: CTYPN
    character(len=5) :: expido, expidb, expidm, insto, instb, instm, celow, cehigh
    character(len=8) :: ctypen(2), ctypem(2), cunito(2), &
         ctypeb(2), cunitb(2),  cunitm(2), ctypeo(2), suffix
    character(len=10) :: obsido, obsidb, obsidm, prefix
    character(len=12) :: estringo, estringb
    character(len=25) :: smodeo, smodeb, smodem, filtero, filterb, filterm, &
         radecsyso, radecsysb, radecsysm
    character(len=60) :: com
    character(len=1024) :: inTEMPIMG_F, inIMAGE_F, inMASKIMG_F, outIMAGE_F
    integer, dimension(:), pointer :: dimVector
    integer :: elowo, elowb, elowm, ehigho, ehighb, ehighm
    integer :: i, ii, ix, iy, j, jj, lun, n, frmtime, type
    integer :: err = 0, typeF, nRowso, nRowsb, nRowsm, typeA
    integer :: revoluto, revolutb, revolutm, isWrong, detloc
    integer :: status = 0, xSizeo, ySizeo, xSizem, ySizem, xSizeb, ySizeb
    integer :: outIMGdim(2)
    integer, parameter :: xGood=780, yGood=780, xDim=900, yDim=900
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: inTEMPIMG_JP, inMASKIMG_JP, &
         inIMAGE_JP, aDI32
    logical :: warndie, die, withmask, withdetxy, withskyxy, i_ok, runesky
    real(kind=single) :: backo, expoo, areao, backb, expob, areab, backm, expom, aream
    real(kind=double) :: crvalo(2), crpixo(2), cdelto(2), crotao(2), &
         crvalb(2), crpixb(2), cdeltb(2), crotab(2), crvalm(2), crpixm(2), &
         cdeltm(2), crotam(2), crotan, rapnto, decpnto, papnto, rapntb, &
         decpntb, papntb, rapntm, decpntm, papntm, ra_nom, dec_nom
    real(kind=double) :: dxcen, dycen
    real(kind=single) :: alpha, beta, rn, skyx, skyy, theta, x, y
    real(kind=single), dimension(:,:), allocatable :: inTEMPIMG, inMASKIMG, inIMAGE, &
         rotarr
    real(double), dimension(:,:), pointer :: aDR64
    real(single), dimension(:,:), pointer :: inTEMPIMG_RP, inIMAGE_RP, inMASKIMG_RP, &
         outIMAGE_RP
    real(single), dimension(:,:), pointer :: aDR32
    ! OLDNEW: OBJFIL = mos1S001-obj-im.fits/mos1S001-fovimsky-750-1250.fits
    ! INBCK  =
    runesky     = .TRUE.
    withmask    = .FALSE.
    withdetxy   = .FALSE.
    withskyxy   = .FALSE.
    i_ok        = .FALSE.
    dxcen = 0.
    dycen = 0.
    skyx = 0.
    skyy = 0.
    ! final, rotated array, future parameter???!!!
    allocate (rotarr(xDim,yDim))
    outIMGdim(1) = xDim
    outIMGdim(2) = yDim
    rotarr = 0.
    ! get the parameters
    inTEMPIMG_F = stringParameter(tempimpar)    ! input template, e.g. mos1S001-fovimsky-750-1250.fits
    inIMAGE_F   = stringParameter(inimpar)      ! input image to rotate, e.g. mos1S001-fovimdet-2000-7200.fits
    withmask    = booleanParameter(withmaskpar) ! apply a mask?
    inMASKIMG_F = ""
    if (withmask) then
       inMASKIMG_F = stringParameter(inmaskpar) ! input mask, e.g. mos1S001-fovimtmask.fits
    endif
    withdetxy = booleanParameter(withdetpar)
    if (withdetxy) then
       dxcen     = realParameter(detxcenpar) ! from esky2det previously run
       dycen     = realParameter(detycenpar) ! from esky2det previously run
       if (dxcen .lt. -99999) then
          call error("NeedDETX","withdetxy=T selected, but no value for detx was entered")
       endif
       if (dycen .lt. -99999) then
          call error("NeedDETY","withdetxy=T selected, but no value for dety was entered")
       endif
       if (trim(adjustl(inTEMPIMG_F)) .ne. 'default') then
          call warning("NoTEMPLATE",&
               "Intemplate incompatable with withdetxy; ignoring intemplate; continuing with DETX/DETY...")
       endif
       runesky = .FALSE.
    else
       if (trim(adjustl(inTEMPIMG_F)) .eq. 'default') then
          call error("NoTemplateFile","If withdetxy=F, you MUST enter a value for intemplate")
       endif
       ! runesky still true from above
    endif
    !
    ! SS: Open the FITS file for the template
    ! open, read in template Int32 image (900,900) and its coordinate keywords
    ! file usually comes from mosspectra/pnspectra
    ! OLDNEW: mos1S001-obj-im-400-750.fits/mos1S001-fovimsky-400-750.fits
    ! OLD: OBJFIL handle NEW: inTEMPIMG_F
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(inTEMPIMG_F),inTEMPIMG_S,1,status,warndie)
    status = 0
    warndie = .TRUE.
    typeF = 3 ! 1 = EVLI, 2 = SPEC, 3 = IMG
    elowo  = 0
    ehigho = 0
    ! SS: Read the FITS file header information
    ! SS: Get the useful keywords
    ! SS: And a few more
    ! BHP: only need some keywords, not actual input array check this???!!!
    call read_header_info(typeF,inTEMPIMG_S,in_T,inTEMPIMG_A,insto,obsido,& 
         expido,smodeo,frmtime,filtero,revoluto,backo,areao,expoo,&
         elowo,ehigho,nRowso,warndie)
    !
    ! get coordinate keywords from template image
    ! ones with suffix 'o' come from the object template image
    !
    warndie = .FALSE.
    call coord_keys (inTEMPIMG_S,rapnto,decpnto,papnto,crvalo,cdelto,&
         cunito,ctypeo,crpixo,crotao,radecsyso,warndie)
    write(errstr,*)'Template crval(1) = ',crvalo(1),' ',' crval(2) = ',crvalo(2)
    call message(AppMsg,Always) ! change to Verbose???!!!
    isWrong = index(ctypeo(1),'DETX')
    if (isWrong .ne. 0) then
       call error("COORDSwrong",&
            "Input template image must be in SKY coordinates, exiting...")
    end if
    call runesky2det(trim(inTEMPIMG_F),insto,crvalo(1),crvalo(2),dxcen,dycen)
    ! don't need template array just xSizeo/ySizeo
    inTEMPIMG_A = array(inTEMPIMG_S, 0, READ)
    dimVector => dimensions(inTEMPIMG_A)
    xSizeo = dimVector(0)
    ySizeo = dimVector(1)
    write(errstr,*) "xSizeo/xDim ySizeo/yDim = ",xSizeo,xDim,ySizeo,yDim
    call message(AppMsg,Noisy)
    !
    if ((xSizeo .ne. xDim) .or. (ySizeo .ne. yDim)) then
       ! should this be a fatal error???!!!
       call error("DIMSwrong","Input template image must be 900x900 Int32")
    end if
    ! user can extract CRPIX1 and CRPIX2 and use those OR input their own values.
    withskyxy = booleanParameter(withskypar)
    if (withskyxy) then 
       skyx  = realParameter(skyxpar)
       skyy  = realParameter(skyypar)
       if ((skyx .lt. -99999) .or. (skyy .lt. -99999)) then ! check this???!!!
          call error("NeedSKYX/Yvalues","withskyxy=T selected, but no values for skyx/skyy were entered")
       endif
    else
       skyx = crpixo(1)
       skyy = crpixo(2)
    endif
    ! why???!!!
    if(skyx .lt. 100.0) skyx = 100.0*skyx                                              
    if(skyy .lt. 100.0) skyy = 100.0*skyy
    !
    ! for constructing energy-added filenames, e.g. mos1S001-bkimsky-350-1000.fits
    ! and other uses down below
    ! MAKE SURE ELOWO and EHIGHO are correct???!!! 10/14/21
    write(celow,"(I5)") elowo
    write(cehigh,"(I5)") ehigho
    if ((elowo .eq. 0) .and. (ehigho .eq. 0)) then
       estringo = ""
    else
       estringo = "-" // trim(adjustl(celow)) // "-" // trim(adjustl(cehigh))
    endif
    !
    select case (insto(1:2))
    case ('M1','M2') ! prefix becomes, e.g. mos1S001 or pnS003
       prefix = "mos" // trim(adjustl(insto(2:2))) // trim(adjustl(expido))
    case ('PN')
       prefix = "pn" // trim(adjustl(expido))
    case default
       call error("INSTwrong",&
            "Instrument must be M1, M2, or PN only, check input file")
    end select
    ! 
    if (withmask) then
       write(errstr,*) "Optional Mask image: ",trim(adjustl(inMASKIMG_F))
       call message(AppMsg,Always)
       ! open, read in MASK image and its coordinate keywords
       status = 0
       warndie = .TRUE.
       call check_open_set(trim(inMASKIMG_F),inMASKIMG_S,1,status,warndie)
       status = 0
       warndie = .TRUE.
       typeF = 3 ! 1 = EVLI, 2 = SPEC, 3 = IMG
       call read_header_info(typeF,inMASKIMG_S,in_T,inMASKIMG_A,instm,obsidm,& 
            expidm,smodem,frmtime,filterm,revolutm,backm,aream,expom,elowm,ehighm,&
            nRowsm,warndie)
       warndie = .FALSE.
       ! don't really need these from the mask, but maybe some sanity checking in
       ! the future???!!!
       call coord_keys (inMASKIMG_S,rapntm,decpntm,papntm,crvalm,cdeltm,&
            cunitm,ctypem,crpixm,crotam,radecsysm,warndie)
       ! check mask array type in ALL CASES???!!!
       call read_2Darray(inMASKIMG_S,inMASKIMG_A,inMASKIMG_RP,aDR64,aDI16,inMASKIMG_JP,&
            xSizem,ySizem,typeA)
       if ((xSizem .ne. xDim) .or. (ySizem .ne. yDim)) then
          call error("DIMSwrong","Input MASK image must be 900x900 Int32")
       end if
       allocate(inMASKIMG(xDim,yDim)) ! 900 x 900 array
       inMASKIMG = 0.                 ! zero it out
       ! mask is 900x900 Int32 array
       ! OLDNEW: mos1S001-mask-im.fits/mos1S001-fovimtmask.fits (or mos1S001-maskimt.fits???)
       ! SS: Read in the standard mask to go with the observation
       !
       ! cast ints as reals for later use, add other types for completeness!!!???
       ! MASK array was only mode=5, no longer used...
       ! OLDNEW: MASKIT/inMASKIMG
       do i=1,xSizem
          do j=1,ySizem
             if (typeA .eq. 2) then ! Int32 array
                inMASKIMG(i,j) = real(inMASKIMG_JP(i-1,j-1)) ! OLDNEW: MASKIT/inMASKIMG
             else ! Real32 array
                inMASKIMG(i,j) = inMASKIMG_RP(i-1,j-1)
             end if
          end do
       end do
       call release(inMASKIMG_S)
    else
       write(errstr,*) "No optional Mask Image selected."
       call message(AppMsg,Always)
    endif
    !
    ! open, read INPUT image file and its coordinate keywords
    !
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(inIMAGE_F),inIMAGE_S,1,status,warndie)
    status = 0
    warndie = .TRUE.
    typeF = 3 ! 1 = EVLI, 2 = SPEC, 3 = IMG
    call read_header_info(typeF,inIMAGE_S,in_T,inIMAGE_A,instb,obsidb,& 
         expidb,smodeb,frmtime,filterb,revolutb,backb,areab,expob,elowb,&
         ehighb,nRowsb,warndie)
    warndie = .FALSE.
    call coord_keys (inIMAGE_S,rapntb,decpntb,papntb,crvalb,cdeltb,&
         cunitb,ctypeb,crpixb,crotab,radecsysb,warndie)
    call read_2Darray(inIMAGE_S,inIMAGE_A,inIMAGE_RP,aDR64,aDI16,inIMAGE_JP,&
         xSizeb,ySizeb,typeA)
    call justincase(inIMAGE_S,inIMAGE_A,ra_nom,dec_nom)
    ! why???!!! the original rot-im-det-sky did this and for some reason CRVAL
    ! in the input images, at least the back-im-det types, CRVAL1/2 == 0.0. why?
    crvalb(1) = ra_nom    ! OLDNEW: CRVALO/crvalb
    crvalb(2) = dec_nom
    ! user-specified, but if still 'default', generate filename.
    ! this works for running rotdet2sky from the command line. In
    ! proton and swcx, rotdet2sky will automatically be called with
    ! outIMG_F set to something correct. I.e. if run standalone,
    ! it will make mos1S001-bkgimsky...fits from keywords, but prot
    ! will create mos1S001-protimsky...fits and swcx will create
    ! mos1S001-swcximsky.fits when they run rotdet2sky within those
    ! tasks. Put this in docs???!!! Fix this permanently???!!!
    !
    write(celow,"(I5)")  elowb
    write(cehigh,"(I5)") ehighb
    if ((elowb .eq. 0) .and. (ehighb .eq. 0)) then
       estringb = ""
    else
       estringb = "-" // trim(adjustl(celow)) // "-" // trim(adjustl(cehigh))
    endif
    !
    write(errstr,*)"Input  IMAGE (DET coords)                = " // trim(adjustl(inIMAGE_F))
    call message(AppMsg,Always)
    outIMAGE_F = stringParameter(outimpar)
    if (trim(adjustl(outIMAGE_F)) .eq. 'default') then
       detloc = index(inIMAGE_F,'det') ! derive it if NOT user-given, make more robust???!!!
       if (detloc .gt. 0) then 
          outIMAGE_F = trim(adjustl(inIMAGE_F(:detloc-1))) // "sky" //  trim(adjustl(estringb)) // ".fits"
       endif
       write(errstr,*)"Output IMAGE (SKY coords, derived)       = " // trim(adjustl(outIMAGE_F))
       call message(AppMsg,Always)
    else
       write(errstr,*)"Output IMAGE (SKY coords, user-selected) = " // trim(adjustl(outIMAGE_F))
       call message(AppMsg,Always)
    end if
    !
    ! SS: Check to see if there is an existing file
    ! BHP: doing this 'early' saves time if clobber is NOT set
    die = .FALSE.
    call check_clobber(outIMAGE_F,die) ! in goflib/gofio
    !
    ! cast ints as reals for later use, add other types for completeness!!!???
    ! this is the image to be rotated
    ! OLDNEW: INBCK/inIMAGE
    allocate(inIMAGE(xDim,yDim)) ! 900 x 900 array, can it be 780x780???!!!
    inIMAGE = 0.                 ! zero it out
    jj = 0
    do i=1,xSizeb
       do j=1,ySizeb
          if (typeA .eq. 2) then ! Int32 array
             inIMAGE(i,j) = real(inIMAGE_JP(i-1,j-1))
          else ! Real32 array
             inIMAGE(i,j) = inIMAGE_RP(i-1,j-1)
          end if
       end do
    end do
    write(errstr,*) "Initial sum of counts in input image: ",sum(inIMAGE)
    call message(AppMsg,Always) ! change to Noisy???!!!
    !
    ! end parameter input section    
    !
    ! now have inTEMPIMG, inIMAGE???!!!, inMASKIMG in 900x900 Real32 arrays
    n = 0
    rn = 0.
    ! Print the details of the template image
    ! print out params
    write(errstr,'(A,1X,A)')'Template image (for coordinate keywords)  :',&
         trim(adjustl(inTEMPIMG_F)) ! OLDNEW:OBJFIL/inTEMPIMG_F
    call message(AppMsg,Always)
    ! BHP: see here:
    ! https://heasarc.gsfc.nasa.gov/docs/software/fitsio/c/f_user/node58.html
    ! it's ALWAYS 8 characters and ftwldp expects C*4 to be one of:
    ! -SIN, -TAN, -ARC, -NCP, -GLS, -MER, and -AIT
    ! cf. https://danmoser.github.io/notes/gai_fits-imgs.html
    !

    CTYPN = ctypeO(1)(5:8)
    !    CTYPN = "-TAN" ! testing ONLY OLDNEW: CTYPN/CTYPN
    write(errstr,'(A,A,1X,A)')'Template image coordinate system   (CTYPE): ',CTYPN !&
    !         trim(adjustl(ctypeo(1))),trim(adjustl(ctypeo(2))) ! CTYPEN/CTYPO
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Template image central coordinates (CRVAL):',&
         crvalo(1)," ",crvalo(2) ! OLDNEW: CRVALN
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Template image reference pixel     (CRPIX):',&
         crpixo(1)," ",crpixo(2) ! OLDNEW: CRPIXN
    call message(AppMsg,Always)
    write(errstr,'(A,E10.4,A,E10.4)')'Template image pixel sizes         (CDELT): ',&
         cdelto(1)," ",cdelto(2) ! OLDNEW: CDELTN
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3)')'Template image rotation angle      (CROTN):',&
         papnto ! PA_PNT converted to CROTN why???!!! OLDNEW: CROTN/papnto
    call message(AppMsg,Always)
    write(errstr,'(A,I5,A,I5)')'Template image nRows               (NAXIS):',&
         xSizeo," ",ySizeo ! OLDNEW: NAXESN(1), NAXESN(2)
    call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)
    !
    ctypeb(1) = ctypeo(1)
    ctypeb(2) = ctypeo(2)
    cdeltb(1) = cdelto(1)
    cdeltb(2) = cdelto(2)
    ctypeo = ctypn
    !
    !print *, "dxcen/dycen = ",dxcen, dycen
    crpixb(1) = (-dxcen + 19499.0)/50.0 - 0.5
    crpixb(2) = (dycen + 19499.0)/50.0 + 0.5
    !
    ! SS: Print the central coords of the observation
    ! BHP: b = O suffixes
    write(errstr,'(A,1X,A)')'Input image to be recast from DET to SKY :',&
         trim(adjustl(inIMAGE_F)) ! OLDNEW: INBCK
    call message(AppMsg,Always)
    write(errstr,'(A,A,1X,A)')'Input image coordinate system   (CTYPE): ', & 
         trim(adjustl(ctypeb(1))),trim(adjustl(ctypeb(2))) ! OLDNEW: CTYPEO
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Input image central coordinates (CRVAL):',&
         crvalb(1)," ",crvalb(2)  ! OLDNEW: CRVALO
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Input image reference pixel     (CRPIX):',&
         crpixb(1)," ",crpixb(2)  ! OLDNEW: CRPIXO
    call message(AppMsg,Always)
    write(errstr,'(A,E10.4,A,E10.4)')'Input image pixel sizes         (CDELT): ',&
         cdeltb(1)," ",cdeltb(2)  ! OLDNEW: CDELTO
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3)')'Input image rotation angle      (CROTN):',&
         papntb
    !    write(errstr,'(A,F8.3,A,F8.3)') 'OV image rotation angle      (CROTN): ',&
    !         crotab(1)," ",crotab(2)
    call message(AppMsg,Always)
    write(errstr,'(A,I5,A,I5)')'Input image nRows               (NAXIS):',&
         xSizeb," ",ySizeb
    call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)
    !
    ! SS:  Read in the image
    ! BHP: entire section in old rot_det_sky_mod.f90 below the SS comment above was
    ! either already done above or (mode 4/5 stuff) been eliminated...
    ! 
    ! original code didn't use any of the mask's coord_keys but sanity check
    ! them here...
    if (withmask) then 
       write(errstr,'(A,1X,A)')'Optional MASK to be applied: ',&
            trim(adjustl(inMASKIMG_F)) ! OLDNEW: MASKIT
       call message(AppMsg,Always)
       write(errstr,'(A,A,1X,A)')'MASK image coordinate system   (CTYPE): ',&
            trim(adjustl(ctypem(1))),trim(adjustl(ctypem(2)))
       call message(AppMsg,Always)
       write(errstr,'(A,F8.3,A,F8.3)')'MASK image central coordinates (CRVAL):',&
            crvalm(1)," ",crvalm(2)
       call message(AppMsg,Always)
       write(errstr,'(A,F8.3,A,F8.3)')'MASK image reference pixel     (CRPIX):',&
            crpixm(1)," ",crpixm(2)
       call message(AppMsg,Always)
       write(errstr,'(A,E10.4,A,E10.4)')'MASK image pixel sizes         (CDELT): ',&
            cdeltm(1)," ",cdeltm(2)
       call message(AppMsg,Always)
       write(errstr,'(A,F8.3)')'MASK image rotation angle      (CROTN):',&
            papntm
       call message(AppMsg,Always)
       write(errstr,'(A,I5,A,I5)')'MASK image nRows               (NAXIS):',&
            xSizeo," ",ySizeo
       call message(AppMsg,Always)
    end if
    ! OLDNEW: crotn ... why does CROTN = PA_PNT in old rot_det_sky???!!!
    crotan = papnto ! mos1 only
    if (insto(1:2) .eq. "M2" ) then 
       crotan = papnto + 90.0  
    elseif (insto(1:2) .eq. "PN" ) then ! PN 
       crotan = papnto - 90.0
    endif
    ! SS: Enter the file to be cast (already done above)
    ! SS: Open the FITS file for the detector coord background image (already done)
    !
    ! BHP: modes eliminated in SAS-19, any input goes, fixed with params. modes 4, 5
    !      also eliminated in SAS-19 (Fall, 2020), mode is 4 taken care of in DAL
    !
    ! BHP: reset the inIMAGE CTYPEB coord keys to those of the template CT
    write(errstr,*)'Converting input DET image coord keywords to template SKY keywords'
    call message(AppMsg,Always)
!
    ctypeb(1) = ctypeo(1)
    ctypeb(2) = ctypeo(2)
    cdeltb(1) = cdelto(1)
    cdeltb(2) = cdelto(2)
    !    ctypo = ctypn never used for anything!

    !    print *, dxcen, dycen
    write(errstr,*)'INIT DXCEN/DYCEN = ',dxcen,dycen
    call message(AppMsg,Always)
    crpixb(1) = (-dxcen + 19499.0)/50.0 - 0.5 ! OLDNEW: CRPIXO
    crpixb(2) = (dycen + 19499.0)/50.0 + 0.5
    ! SS: Cast the new observation into the mosaic
    ! if (mode .le. 4) then... section
    ! OLDNEW: CTYPEO
    dxcen = (dxcen + 19499.0)/50.0 + 0.5 ! detx param
    dycen = (dycen + 19499.0)/50.0 + 0.5 ! dety param
    theta = -0.01745329 * crotan ! OLDNEW: crotn
    write(errstr,*)'NEW DXCEN/DYCEN  = ',dxcen,dycen
    call message(AppMsg,Always)

    alpha = skyx + dxcen * cos(theta) - dycen * sin(theta)
    beta  = skyy - dycen * cos(theta) - dxcen * sin(theta)
    write(errstr,*)'alpha/beta = ',alpha,beta
    call message(AppMsg,Verbose) ! change to Noisy???!!!

    x = alpha - dxcen * cos(theta) + dycen * sin(theta)
    y = beta  + dycen * cos(theta) + dxcen * sin(theta)
    write(errstr,*)'x/y = ',x,y
    call message(AppMsg,Verbose) ! change to Noisy???!!!

    do i=1,xGood     ! 780 check this for ALL CASES???!!!
       do ii=1,yGood ! 780 check this for ALL CASES???!!!
          if(inIMAGE(i,ii) .ne. 0.) then  ! was: input(780,780)
             x = alpha - i * cos(theta) + ii * sin(theta)
             y = beta + ii * cos(theta) + i * sin(theta)
             ix = nint(x)
             iy = nint(y)
             if((ix .ge. 1) .and. (ix .le. xDim) .and. & ! OLDNEW: hardcoded 900
                  (iy .ge. 1) .and. (iy .le. yDim)) then ! OLDNEW: hardcoded 900
                rotarr(ix,iy) = rotarr(ix,iy) + inIMAGE(i,ii) ! OLDNEW: array 'input'
                n = n + 1
                rn = rn + inIMAGE(i,ii)
             endif
          endif
       enddo
    enddo
    write(errstr,*) 'Non-zero input pixels = ',n
    call message(AppMsg,Always)
    write(errstr,*) 'Cast counts           = ',rn
    call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)

    !  Do the standard masking

    if (withmask) then
       n = 0
       rn = 0.0
       do i=1,xDim      ! 900
          do ii=1,yDim  ! 900
             if(inMASKIMG(i,ii) .eq. 0) then
                rotarr(i,ii) = 0.
             else
                n = n + 1
                rn = rn + rotarr(i,ii)
             endif
          enddo
       enddo
       !
       write(errstr,*) 'Values after standard masking:'
       call message(AppMsg,Always)
       write(errstr,*) 'Non-zero input pixels = ',n
       call message(AppMsg,Always)
       write(errstr,*) 'Cast counts           = ',rn
    endif
    !  Print the details of the output image
    write(errstr,'(A,1X,A)')'Final (rotated) image in SKY coords        : ',&
         trim(adjustl(outIMAGE_F))
    call message(AppMsg,Always)
    write(errstr,'(A,A,1X,A)')'Final SKY image coordinate system   (CTYPE): ',&
         trim(adjustl(ctypeo(1))),trim(adjustl(ctypeo(2)))
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Final SKY image central coordinates (CRVAL):',&
         crvalo(1)," ",crvalo(2)
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Final SKY image reference pixel     (CRPIX):',&
         crpixo(1)," ",crpixo(2)
    call message(AppMsg,Always)
    write(errstr,'(A,E10.4,A,E10.4)')'Final SKY image pixel sizes         (CDELT): ',&
         cdelto(1)," ",cdelto(2)
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3)')'Final SKY image rotation angle      (CROTN):',&
         papnto
    call message(AppMsg,Always)
    write(errstr,'(A,I5,A,I5)')'Final SKY image nRows               (NAXIS):',&
         xSizeo," ",ySizeo
    call message(AppMsg,Always)
    write(errstr,*)""
    call message(AppMsg,Always)  
    !
    ! open output image file outIMAGE_F (unless clobber)
    !
    ! open output SKY coord background map
    ! open output image file outIMAGE_F
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_open_set(outIMAGE_F,outIMAGE_S,3,status,warndie) ! 3=CREATE, goflib/gofio
    outIMAGE_A = addArray(outIMAGE_S,"PRIMARY",real32,outIMGdim)
    outIMAGE_RP => real32Array2Data(outIMAGE_A) ! 'final product'
    !
    ! SS: Write some useful header keywords
    !
    write(errstr,*) "Copying some attributes from input template image to output image"
    call message(AppMsg,Always)
    call copyAllAttributes(attributable(inTEMPIMG_S),attributable(outIMAGE_S))
    call copyAllAttributes(attributable(inTEMPIMG_A),attributable(outIMAGE_A))
    ! here we clone the template image (900,900) 
    !write(errstr,*) "Cloning Template IMAGE to Output Image..."
    !call message(AppMsg,Always)  
    !outIMAGE_S = clone(trim(adjustl(inTEMPIMG_F)),trim(adjustl(outIMAGE_F)),MODIFY)
    ! then we work on IT, recasting it's array values to rotarr from above
    !call read_2Darray(outIMAGE_S,outIMAGE_A,outIMAGE_RP,aDR64,aDI16,aDI32,&
    !     xSizeo,ySizeo,typeA)
    write(errstr,*) "Filling new output IMAGE array with rotated values..."
    call message(AppMsg,Always)
    !
    do i=0,xSizeo-1
       do j=0,ySizeo-1
          outIMAGE_RP(i,j) = rotarr(i+1,j+1)
       end do
    end do
    !
    write(errstr,*) "Releasing (closing) output IMAGE file."
    call message(AppMsg,Always)
    
    call release(inTEMPIMG_S) ! close template image
    call release(outIMAGE_S)  ! close output image
    
    return

  end subroutine rotdet2sky

  subroutine justincase(in_S,in_A,ra,dec)

    type(DataSetT) :: in_S
    type(ArrayT)   :: in_A
    real(kind=double) :: ra, dec
    
    if (hasAttribute(in_A,"RA_NOM")) then
       ra = real64Attribute(in_A,"RA_NOM")
    elseif (hasAttribute(in_S,"RA_NOM")) then
       ra = real64Attribute(in_S,"RA_NOM")
    else
       call error ("imgNoRA_NOM","File does not have RA_NOM attribute")
    end if
    if (hasAttribute(in_A,"DEC_NOM")) then
       dec = real64Attribute(in_A,"DEC_NOM")
    elseif (hasAttribute(in_S,"DEC_NOM")) then
       dec = real64Attribute(in_S,"DEC_NOM")
    else
       call error ("imgNoRA_NOM","File does not have RA_NOM attribute")
    end if
    
    return
    
  end subroutine justincase
  
end module rotdet2sky_mod
